from collections.abc import Sequence
from datetime import timedelta

import httpx
from pydantic import TypeAdapter

from camunda_client.clients.dto import AuthData
from camunda_client.clients.endpoints import CamundaUrls
from camunda_client.types_ import Variables
from camunda_client.utils import camunda_timedelta, raise_for_status

from .dto import ExternalTaskConfig
from .schemas import (
    CompleteExternalTaskSchema,
    ExtendLockOnExternalTaskSchema,
    ExternalTaskFailureSchema,
    FetchExternalTasksSchema,
    FetchExternalTaskTopicSchema,
    TaskBpmnErrorSchema,
)
from .schemas.response import ExternalTaskSchema


class ExternalTaskClient:
    def __init__(  # noqa: PLR0913
        self,
        worker_id: str,
        base_url: str,
        auth_data: AuthData,
        transport: httpx.AsyncHTTPTransport,
        config: ExternalTaskConfig | None = None,
        urls: CamundaUrls | None = None,
    ) -> None:
        self._worker_id = worker_id
        self._http_client = httpx.AsyncClient(
            base_url=base_url,
            transport=transport,
            auth=httpx.BasicAuth(
                username=auth_data.username,
                password=auth_data.password,
            ),
            headers={"Content-Type": "application/json"},
        )
        self._urls = urls or CamundaUrls()
        self._config = config or ExternalTaskConfig()

    async def fetch_and_lock(
        self,
        topic_names: Sequence[str],
        business_key: str | None = None,
        process_variables: Variables | None = None,
        lock_timeout: timedelta | None = None,
    ) -> Sequence[ExternalTaskSchema]:
        url = self._urls.external_task.fetch_and_lock
        topics = [
            FetchExternalTaskTopicSchema(
                topic_name=topic_name,
                lock_duration=camunda_timedelta(self._config.lock_duration),
                process_variables=process_variables,
                business_key=business_key,
                variables=None,
                has_local_variables=None,
            )
            for topic_name in topic_names
        ]
        schema = FetchExternalTasksSchema(
            worker_id=self._worker_id,
            max_tasks=self._config.max_tasks,
            lock_timeout=camunda_timedelta(
                lock_timeout or self._config.async_response_timeout,
            ),
            topics=topics,
        )
        try:
            response = await self._http_client.post(
                url,
                content=schema.model_dump_json(
                    by_alias=True,
                    exclude_none=True,
                ),
                timeout=60,
            )
        except httpx.ReadTimeout:
            return []

        raise_for_status(response)
        adapter = TypeAdapter(list[ExternalTaskSchema])
        return adapter.validate_python(response.json())

    async def complete(
        self,
        task_id: str,
        *,
        global_variables: Variables | None = None,
        local_variables: Variables | None = None,
    ) -> None:
        url = self._urls.external_task.complete(task_id)
        schema = CompleteExternalTaskSchema(
            worker_id=self._worker_id,
            variables=global_variables,
            local_variables=local_variables,
        )
        content = schema.model_dump_json(
            by_alias=True,
            exclude_none=True,
        )

        response = await self._http_client.post(url, content=content)
        raise_for_status(response)

    async def failure(
        self,
        task_id: str,
        *,
        error_message: str,
        retries: int | None = None,
        error_details: str | None = None,
    ) -> None:
        url = self._urls.external_task.failure(task_id)
        schema = ExternalTaskFailureSchema(
            worker_id=self._worker_id,
            error_message=error_message,
            retries=retries,
            retry_timeout=camunda_timedelta(self._config.retry_timeout),
            error_details=error_details,
        )

        response = await self._http_client.post(
            url,
            content=schema.model_dump_json(
                by_alias=True,
                exclude_none=True,
            ),
        )
        raise_for_status(response)

    async def extend_lock(
        self,
        task_id: str,
        *,
        new_duration: int | None = None,
    ) -> None:
        url = self._urls.external_task.extend_lock(task_id)
        schema = ExtendLockOnExternalTaskSchema(
            worker_id=self._worker_id,
            new_duration=new_duration or camunda_timedelta(self._config.lock_duration),
        )

        response = await self._http_client.post(
            url,
            content=schema.model_dump_json(
                by_alias=True,
                exclude_none=True,
            ),
        )
        raise_for_status(response)

    async def unlock(
        self,
        task_id: str,
    ) -> None:
        url = self._urls.external_task.unlock(task_id)
        response = await self._http_client.post(url)
        raise_for_status(response)

    async def bpmn_error(
        self,
        task_id: str,
        error_code: str,
        error_message: str,
    ) -> None:
        url = self._urls.external_task.bpmn_error(task_id)
        schema = TaskBpmnErrorSchema(
            error_code=error_code,
            error_message=error_message,
        )
        response = await self._http_client.post(
            url,
            content=schema.model_dump_json(
                by_alias=True,
                exclude_none=True,
            ),
        )
        raise_for_status(response)



import json
from collections.abc import Sequence
from http import HTTPStatus
from uuid import UUID

import httpx
from pydantic.type_adapter import TypeAdapter

from camunda_client.clients.dto import AuthData
from camunda_client.clients.endpoints import CamundaUrls
from camunda_client.clients.engine.schemas.body import ClaimTaskSchema
from camunda_client.clients.engine.schemas.response import (
    HistoricTaskInstanceSchema,
    TaskSchema,
    VariableInstanceSchema,
)
from camunda_client.clients.schemas import PaginationParams
from camunda_client.types_ import Variables
from camunda_client.utils import raise_for_status

from .schemas import (
    GetHistoryTasksFilterSchema,
    GetTasksFilterSchema,
    ProcessInstanceQuerySchema,
    ProcessInstanceSchema,
    StartProcessInstanceSchema,
)


class CamundaEngineClient:
    def __init__(
        self,
        base_url: str,
        auth_data: AuthData,
        transport: httpx.AsyncHTTPTransport,
        urls: CamundaUrls | None = None,
    ) -> None:
        self._http_client = httpx.AsyncClient(
            base_url=base_url,
            transport=transport,
            auth=httpx.BasicAuth(
                username=auth_data.username,
                password=auth_data.password,
            ),
            headers={"Content-Type": "application/json"},
        )
        self._urls = urls or CamundaUrls()

    async def start_process(
        self,
        process_key: str,
        business_key: str | None = None,
        variables: Variables | None = None,
        tenant_id: str | None = None,
    ) -> ProcessInstanceSchema:
        """
        Instantiates a given process definition, starts the latest
        version of the process definition which belongs to no tenant.
        Process variables and business key may be supplied in the request body
        """

        url = self._urls.get_start_process_instance(process_key, tenant_id)
        schema = StartProcessInstanceSchema(
            business_key=business_key,
            variables=variables,
        )
        content = schema.model_dump_json(by_alias=True)
        response = await self._http_client.post(url, content=content)
        raise_for_status(response)
        return ProcessInstanceSchema.model_validate(response.json())

    async def get_process_instances(
        self,
        params: ProcessInstanceQuerySchema,
    ) -> Sequence[ProcessInstanceSchema]:
        """
        Queries for process instances that fulfill given parameters.
        Parameters may be static as well as dynamic runtime properties
        of process instances. The size of the result set can be
        retrieved by using the Get Instance Count method
        """

        response = await self._http_client.get(
            self._urls.process_instances,
            params=params.model_dump(
                mode="json",
                by_alias=True,
                exclude_none=True,
            ),
        )
        raise_for_status(response)
        adapter = TypeAdapter(list[ProcessInstanceSchema])
        return adapter.validate_python(response.json())

    async def delete_process(self, process_instance_id: str) -> None:
        """Deletes a running process instance by id"""

        url = self._urls.get_process_instance(process_instance_id)
        response = await self._http_client.delete(
            url,
            params={"skipCustomListeners": "true", "skipIoMappings": "true"},
        )
        raise_for_status(response)

    async def get_tasks(
        self,
        schema: GetTasksFilterSchema,
        pagination: PaginationParams | None = None,
    ) -> Sequence[TaskSchema]:
        """Queries for tasks that fulfill a given filter"""

        pagination_params = (
            pagination.model_dump(mode="json", by_alias=True) if pagination else {}
        )

        response = await self._http_client.post(
            self._urls.task,
            params=pagination_params,
            content=schema.model_dump_json(by_alias=True, exclude_none=True),
        )
        raise_for_status(response)
        adapter = TypeAdapter(list[TaskSchema])
        return adapter.validate_python(response.json())

    async def get_task(
        self,
        ident: UUID,
    ) -> TaskSchema | None:
        """Retrieves a task by id"""

        url = self._urls.get_task_by_id(str(ident))
        response = await self._http_client.get(url)

        if response.status_code == HTTPStatus.NOT_FOUND:
            return None

        raise_for_status(response)
        return TaskSchema.model_validate(response.json())

    async def get_history_tasks(
        self,
        schema: GetHistoryTasksFilterSchema,
        pagination: PaginationParams | None = None,
    ) -> Sequence[HistoricTaskInstanceSchema]:
        pagination_params = (
            pagination.model_dump(mode="json", by_alias=True) if pagination else {}
        )

        response = await self._http_client.post(
            self._urls.history_task,
            params=pagination_params,
            content=schema.model_dump_json(by_alias=True),
        )
        raise_for_status(response)
        adapter = TypeAdapter(list[HistoricTaskInstanceSchema])
        return adapter.validate_python(response.json())

    async def get_variable_instances(
        self,
        *,
        process_instance_id: UUID,
        deserialize_values: bool = False,
    ) -> Sequence[VariableInstanceSchema]:
        response = await self._http_client.get(
            self._urls.variable_instances,
            params={
                "processInstanceId": str(process_instance_id),
                "deserializeValues": deserialize_values,
            },
        )
        raise_for_status(response)
        adapter = TypeAdapter(list[VariableInstanceSchema])
        return adapter.validate_python(response.json())

    async def submit_task_form(
        self,
        task_id: UUID,
        variables: Variables | None = None,
    ) -> None:
        """
        Completes a task and updates process variables using a form submit.
        """
        variables = variables or {}
        content = json.dumps(
            {
                "variables": {
                    key: value.model_dump(mode="json", by_alias=True)
                    for key, value in variables.items()
                },
            },
        )
        response = await self._http_client.post(
            self._urls.submit_task_form(str(task_id)),
            content=content,
        )
        raise_for_status(response)

    async def claim_task(
        self,
        task_id: UUID,
        user_id: UUID,
    ) -> None:
        """
        Claims a task for a specific user.
        """
        response = await self._http_client.post(
            self._urls.claim_task(str(task_id)),
            content=ClaimTaskSchema(user_id=str(user_id)).model_dump_json(
                by_alias=True,
            ),
        )
        raise_for_status(response)

    async def unclaim_task(
        self,
        task_id: UUID,
    ) -> None:
        """
        Claims a task for a specific user.
        """
        response = await self._http_client.post(self._urls.unclaim_task(str(task_id)))
        raise_for_status(response)


# • эндпоинт старта процесса
# • эндпоинт получения списка юзер таск
# эндпоинт получения списка юзер таск через фильтры ?
# • эндпоинт для получения переменных user_task get_variables
# • эндпоинт для завершения task
# • эндпоинт get task history
# • POST /task/{id}/claim
# • POST /task/anId/unclaim


"""


    async def get_user_tasks(self) -> dict[str, Any]:
        raise NotImplementedError

    async def get_variables(self) -> dict[str, Any]:
        raise NotImplementedError

    async def get_process_history(self) -> dict[str, Any]:
        raise NotImplementedError

    async def get_task_count(self, user_id: UUID) -> dict[str, Any]:
        raise NotImplementedError


"""



from datetime import datetime
from uuid import UUID

from camunda_client.types_ import (
    BaseSchema,
    MayBeNullableList,
    Variables,
    VariableValueSchema,
)

from .enums import DelegationState


class LinkSchema(BaseSchema):
    rel: str | None = None
    href: str | None = None
    method: str | None = None


class ProcessInstanceSchema(BaseSchema):
    id: UUID
    definition_id: str | None = None
    links: MayBeNullableList[LinkSchema]
    business_key: str | None = None
    case_instance_id: str | None = None
    suspended: bool | None = None
    tenant_id: str | None = None


class ProcessDefinitionSchema(BaseSchema):
    id: str
    key: str | None = None
    category: str | None = None
    description: str | None = None
    name: str | None = None
    version: int | None = None
    resource: str | None = None
    deployment_id: str | None = None
    diagram: str | None = None
    suspended: bool | None = None
    tenant_id: str | None = None
    version_tag: str | None = None
    history_time_to_live: str | None = None
    startable_in_task_list: bool | None = None
    variables: Variables


class TaskSchema(BaseSchema):
    id: UUID
    name: str
    assignee: str
    owner: str | None = None
    created: datetime
    last_updated: datetime | None
    delegation_state: DelegationState | None
    description: str | None
    execution_id: UUID
    parent_task_id: str | None
    priority: int
    process_definition_id: str
    process_instance_id: UUID
    case_execution_id: str | None = None
    case_definition_id: str | None = None
    case_instance_id: str | None = None
    task_definition_key: str | None = None
    suspended: bool
    tenant_id: str | None


class HistoricTaskInstanceSchema(BaseSchema):
    id: UUID
    process_definition_key: str
    process_definition_id: str
    process_instance_id: UUID
    execution_id: str
    case_definition_key: str | None = None
    case_definition_id: str | None = None
    case_instance_id: str | None = None
    case_execution_id: str | None = None
    activity_instance_id: str | None = None
    name: str | None = None
    description: str | None = None
    delete_reason: str | None = None
    owner: str | None = None
    assignee: str | None = None
    start_time: datetime
    end_time: datetime
    duration: int
    task_definition_key: str
    priority: int | None = None
    due: datetime | None = None
    parent_task_id: str | None = None
    follow_up: datetime | None = None
    tenant_id: str | None = None
    removal_time: datetime | None = None
    root_process_instance_id: UUID | None = None


class VariableInstanceSchema(VariableValueSchema):
    id: str
    name: str
    process_definition_id: str
    process_instance_id: UUID
    execution_id: str
    case_instance_id: str | None = None
    case_execution_id: str | None = None
    task_id: str | None = None
    batch_id: str | None = None
    activity_instance_id: str | None = None
    tenant_id: str | None = None
    error_message: str | None = None



from datetime import datetime, timedelta
from decimal import Decimal
from typing import Any

import httpx
import orjson

from camunda_client.exceptions import CamundaClientError
from camunda_client.types_ import VariableTypes, VariableValueSchema


def raise_for_status(response: httpx.Response) -> None:
    if response.is_success:
        return

    raise CamundaClientError(
        status_code=response.status_code,
        response_data=response.content,
    )


def camunda_timedelta(delta: timedelta) -> int:
    return int(delta.total_seconds() * 1000)


def deserialize(
    value: Any,  # noqa: ANN401
    type_: VariableTypes | None = None,
) -> VariableValueSchema:
    if type_ is not None:
        return VariableValueSchema(value=value, type=type_)

    match value:
        case bool():
            type_ = "Boolean"
        case datetime():
            type_ = "Date"
            value = to_camunda_datetime(value)
        case Decimal():
            type_ = "Double"
        case float():
            type_ = "Double"
        case int():
            type_ = "Integer"
        case None:
            type_ = "Null"
        case dict() | list():
            type_ = "Json"
            value = orjson.dumps(value)
        case bytes():
            type_ = "Bytes"
        case str():
            type_ = "String"
        case _ as never:
            msg = f"Got undefined type: {type(never)}"
            raise ValueError(msg)

    return VariableValueSchema(value=value, type=type_)


def to_camunda_datetime(value: datetime) -> str:
    return value.isoformat(timespec="milliseconds").replace("+00:00", "+0000")


def process_variable(variable: VariableValueSchema) -> Any:  # noqa: ANN401
    if variable.type == "Json":
        return orjson.loads(variable.value)

    return variable.value



from typing import ClassVar


class ExternalTaskUrls:
    _base_path: ClassVar[str] = "/external-task"

    fetch_and_lock = f"{_base_path}/fetchAndLock"

    @classmethod
    def complete(cls, ident: str) -> str:
        return f"{cls._base_path}/{ident}/complete"

    @classmethod
    def failure(cls, ident: str) -> str:
        return f"{cls._base_path}/{ident}/failure"

    @classmethod
    def extend_lock(cls, ident: str) -> str:
        return f"{cls._base_path}/{ident}/extendLock"

    @classmethod
    def unlock(cls, ident: str) -> str:
        return f"{cls._base_path}/{ident}/unlock"

    @classmethod
    def bpmn_error(cls, ident: str) -> str:
        return f"{cls._base_path}/{ident}/bpmnError"


class CamundaUrls:
    external_task = ExternalTaskUrls()

    deployment_create = "/deployment/create"
    message_send = "/message"
    process_instances = "/process-instance"
    task = "/task"
    history_task = "/history/task"
    variable_instances = "/history/variable-instance"

    @staticmethod
    def get_start_process_instance(
        process_key: str,
        tenant_id: str | None = None,
    ) -> str:
        if tenant_id:
            return f"/process-definition/key/{process_key}/tenant-id/{tenant_id}/start"
        return f"/process-definition/key/{process_key}/start"

    @staticmethod
    def get_process_instance(process_instance_id: str) -> str:
        return f"/process-instance/{process_instance_id}"

    @classmethod
    def get_task_by_id(cls, ident: str) -> str:
        return f"{cls.task}/{ident}"

    @classmethod
    def submit_task_form(cls, task_id: str) -> str:
        return f"{cls.task}/{task_id}/submit-form"

    @classmethod
    def claim_task(cls, task_id: str) -> str:
        return f"{cls.task}/{task_id}/claim"

    @classmethod
    def unclaim_task(cls, task_id: str) -> str:
        return f"{cls.task}/{task_id}/unclaim"



from uuid import UUID

from camunda_client.types_ import BaseSchema, Variables

from .enums import SortOrder, TaskQuerySortEnum


class StartProcessInstanceSchema(BaseSchema):
    business_key: str | None = None
    variables: Variables | None = None


class ClaimTaskSchema(BaseSchema):
    user_id: str


class SortSchema(BaseSchema):
    sort_by: TaskQuerySortEnum | None = None
    sort_order: SortOrder | None = None


class GetTasksFilterSchema(BaseSchema):
    assignee: str | None = None
    candidate_group: str | None = None
    candidate_user: str | None = None
    sorting: SortSchema | None = None
    process_instance_id: UUID | None = None


class GetHistoryTasksFilterSchema(BaseSchema):
    process_instance_id: UUID | None = None
    sorting: SortSchema | None = None



from typing import Any

import pydantic

from camunda_client.types_ import BaseSchema, Variables


class FetchExternalTaskTopicSchema(BaseSchema):
    topic_name: str
    lock_duration: int
    variables: list[str] | None = None
    has_local_variables: bool | None = pydantic.Field(None, alias="localVariables")
    business_key: str | None = None
    process_definition_id: str | None = None
    process_definition_id_in: list[str] | None = None
    process_definition_key: str | None = None
    process_definition_key_in: list[str] | None = None
    process_definition_version_tag: str | None = None
    without_tenant_id: bool | None = None
    tenant_id_in: list[str] | None = None
    process_variables: dict[str, Any] | None = None
    deserialize_values: bool | None = None
    include_extension_properties: bool | None = None


class FetchExternalTasksSchema(BaseSchema):
    worker_id: str
    max_tasks: int
    use_priority: bool | None = None
    lock_timeout: int | None = pydantic.Field(None, alias="asyncResponseTimeout")
    topics: list[FetchExternalTaskTopicSchema] | None = None


class CompleteExternalTaskSchema(BaseSchema):
    worker_id: str
    variables: Variables | None = None
    local_variables: Variables | None = None


class ExternalTaskFailureSchema(BaseSchema):
    worker_id: str
    error_message: str
    error_details: str | None = None
    retries: int | None = None
    retry_timeout: int | None = None
    variables: Variables | None = None
    local_variables: Variables | None = None


class ExtendLockOnExternalTaskSchema(BaseSchema):
    worker_id: str
    new_duration: int


class TaskBpmnErrorSchema(BaseSchema):
    error_code: str
    error_message: str | None
    variables: Variables | None = None



import asyncio
import contextlib
from collections.abc import AsyncIterator
from datetime import timedelta
from types import TracebackType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from camunda_client.clients import ExternalTaskClient

from .context import ExternalTaskContext
from .dto import ExternalTaskDTO
from .topic_consumer import TopicConsumer


class ExternalTaskWorker:
    def __init__(
        self,
        client: "ExternalTaskClient",
        pull_interval: timedelta,
    ) -> None:
        self._consumers: dict[str, TopicConsumer] = {}
        self._pull_interval = pull_interval
        self._client = client

        self._tg = asyncio.TaskGroup()
        self._closing = asyncio.Event()

    async def __aenter__(self) -> None:
        await self._tg.__aenter__()
        self._tg.create_task(self._pull_tasks())

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._closing.set()
        await self._tg.__aexit__(exc_type, exc_val, exc_tb)

    async def _pull_tasks(self) -> None:
        while True:
            if self._consumers:
                tasks = await self._client.fetch_and_lock(
                    topic_names=list(self._consumers),
                )

                for task in tasks:
                    consumer = self._consumers[task.topic_name]
                    consumer.task_contexts.append(
                        ExternalTaskContext(
                            client=self._client,
                            task=ExternalTaskDTO.model_validate(task),
                        ),
                    )
                    consumer.new_task_event.set()

            await self._wait()

            if self._closing.is_set():
                return

    async def _wait(self) -> None:
        _, pending = await asyncio.wait(
            [
                asyncio.create_task(
                    asyncio.sleep(self._pull_interval.total_seconds()),
                ),
                asyncio.create_task(self._closing.wait()),
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in pending:
            task.cancel()

    @contextlib.asynccontextmanager
    async def subscribe(
        self,
        topic: str,
    ) -> AsyncIterator[AsyncIterator[ExternalTaskContext]]:
        topic_consumer = TopicConsumer(closing=self._closing)

        if topic in self._consumers:
            raise ValueError

        self._consumers[topic] = topic_consumer

        yield topic_consumer

        del self._consumers[topic]



from types import TracebackType
from typing import TYPE_CHECKING

from camunda_client.exceptions import InvalidStateError

if TYPE_CHECKING:
    from camunda_client.clients import ExternalTaskClient
from camunda_client.types_ import Variables

from .dto import ExternalTaskDTO


class ExternalTaskContext:
    def __init__(
        self,
        client: "ExternalTaskClient",
        task: ExternalTaskDTO,
    ) -> None:
        self._client = client
        self._task = task
        self._closed: bool = False

    async def __aenter__(self) -> ExternalTaskDTO:
        return self._task

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._closed:
            return

        if exc_val:
            error_message = f"{exc_val.__class__.__name__}"
            await self._client.failure(self._task.id, error_message=error_message)
        else:
            await self._client.complete(self._task.id)

    async def unlock_task(self) -> None:
        await self._client.unlock(self._task.id)

    async def complete(
        self,
        global_variables: Variables | None = None,
        local_variables: Variables | None = None,
    ) -> None:
        if self._closed:
            msg = "TaskContext is already closed"
            raise InvalidStateError(msg)

        await self._client.complete(
            self._task.id,
            global_variables=global_variables,
            local_variables=local_variables,
        )
        self._closed = True

    async def fail(
        self,
        error_message: str,
        retries: int | None = None,
        error_details: str | None = None,
    ) -> None:
        if self._closed:
            msg = "TaskContext is already closed"
            raise InvalidStateError(msg)

        await self._client.failure(
            self._task.id,
            error_message=error_message,
            retries=retries,
            error_details=error_details,
        )
        self._closed = True



from enum import Enum


class ProcessInstanceQuerySortEnum(Enum):
    INSTANCE_ID = "instanceId"
    DEFINITION_ID = "definitionId"
    DEFINITION_KEY = "definitionKey"
    BUSINESS_KEY = "businessKey"
    TENANT_ID = "tenantId"


class TaskQuerySortEnum(Enum):
    ASSIGNEE = "assignee"
    CASE_EXECUTION_ID = "caseExecutionId"
    CASE_EXECUTION_VARIABLE = "caseExecutionVariable"
    CASE_INSTANCE_ID = "caseInstanceId"
    CASE_INSTANCE_VARIABLE = "caseInstanceVariable"
    CREATED = "created"
    DESCRIPTION = "description"
    DUE_DATE = "dueDate"
    EXECUTION_ID = "executionId"
    EXECUTION_VARIABLE = "executionVariable"
    FOLLOW_UP_DATE = "followUpDate"
    ID = "id"
    INSTANCE_ID = "instanceId"
    LAST_UPDATED = "lastUpdated"
    NAME = "name"
    NAME_CASE_INSENSITIVE = "nameCaseInsensitive"
    PRIORITY = "priority"
    PROCESS_VARIABLE = "processVariable"
    TASK_VARIABLE = "taskVariable"


class SortOrder(Enum):
    ASC = "asc"
    DESC = "desc"


class Operator(Enum):
    EQ = "eq"
    NEQ = "neq"
    GT = "gt"
    GTEQ = "gteq"
    LT = "lt"
    LTEQ = "lteq"
    LIKE = "like"
    NOT_LIKE = "notLike"


class DelegationState(Enum):
    PENDING = "PENDING"
    RESOLVED = "RESOLVED"



from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True, slots=True)
class CamundaClientError(Exception):
    status_code: int
    response_data: Any = None

    def __str__(self) -> str:
        return f"HTTP CODE: {self.status_code}, RESPONSE DATA: {self.response_data}"


class InvalidStateError(Exception):
    pass



from .body import (
    GetHistoryTasksFilterSchema,
    GetTasksFilterSchema,
    StartProcessInstanceSchema,
)
from .query import ProcessInstanceQuerySchema
from .response import ProcessInstanceSchema

__all__ = [
    "StartProcessInstanceSchema",
    "ProcessInstanceQuerySchema",
    "GetHistoryTasksFilterSchema",
    "ProcessInstanceSchema",
    "GetTasksFilterSchema",
]



from datetime import datetime
from typing import Any

from camunda_client.types_ import BaseDTO, TVariables, Variables


class ExternalTaskDTO(BaseDTO):
    id: str
    worker_id: str
    topic_name: str
    activity_id: str | None = None
    activity_instance_id: str | None = None
    error_message: str | None = None
    execution_id: str | None = None
    lock_expiration_time: datetime | None = None
    process_definition_id: str | None = None
    process_definition_key: str | None = None
    process_instance_id: str | None = None
    tenant_id: str | None = None
    retries: int | None = None
    business_key: str | None = None
    variables: Variables
    parsed_variables: dict[str, Any]

    def get_variables(self, type_: type[TVariables]) -> TVariables:
        return type_.model_validate(self.parsed_variables)



from datetime import datetime
from typing import Any

from camunda_client.types_ import BaseSchema, Variables
from camunda_client.utils import process_variable


class ExternalTaskSchema(BaseSchema):
    id: str
    worker_id: str
    topic_name: str
    activity_id: str | None = None
    activity_instance_id: str | None = None
    error_message: str | None = None
    execution_id: str | None = None
    lock_expiration_time: datetime | None = None
    process_definition_id: str | None = None
    process_definition_key: str | None = None
    process_definition_version_tag: str | None = None
    process_instance_id: str | None = None
    tenant_id: str | None = None
    retries: int | None = None
    suspended: bool | None = None
    priority: int | None = None
    business_key: str | None = None
    variables: Variables

    @property
    def parsed_variables(self) -> dict[str, Any]:
        return {key: process_variable(schema) for key, schema in self.variables.items()}
